#include "postgres.h"

#include <math.h>

#include "access/heapam.h"
#include "access/htup_details.h"
#include "access/nbtree.h"
#include "access/transam.h"
#include "access/xact.h"
#include "catalog/index.h"
#include "catalog/indexing.h"
#include "catalog/pg_am.h"
#include "commands/tablecmds.h"
#include "executor/spi.h"
#include "lib/bloomfilter.h"
#include "miscadmin.h"
#include "nodes/makefuncs.h"
#include "storage/lmgr.h"
#include "utils/memutils.h"
#include "utils/snapmgr.h"
#include "utils/fmgroids.h"


PG_MODULE_MAGIC;

PG_FUNCTION_INFO_V1(test_split);


/* Insert new test tuple. Returns the TID */
static ItemPointerData
insert_tuple(Relation indrel, Relation heaprel, char *tuple)
{
	HeapTupleData htup;
	HeapTupleHeader td;

	td = DatumGetHeapTupleHeader(OidInputFunctionCall(F_RECORD_IN, tuple,
													  heaprel->rd_rel->reltype, -1));
	htup.t_len = HeapTupleHeaderGetDatumLength(td);
	ItemPointerSetInvalid(&htup.t_self);
	htup.t_tableOid = RelationGetRelid(heaprel);
	htup.t_data = td;

	CatalogTupleInsert(heaprel, &htup);
	
	return htup.t_self;
}

static char *
print_index_tuple(Relation indrel, IndexTuple itup)
{
	StringInfoData sbuf;
	int			natts;
	AttrNumber	attno;

	initStringInfo(&sbuf);

	natts = IndexRelationGetNumberOfAttributes(indrel);
	for (attno = 1; attno <= natts; attno++)
	{
		Datum		d;
		bool		isnull;

		d = index_getattr(itup, attno, RelationGetDescr(indrel), &isnull);

		appendStringInfo(&sbuf, "%d ", DatumGetInt32(d));
	}

	return sbuf.data;
}

static void
print_split(Relation indrel, ItemPointerData split_tid, char *title)
{
	/* scan the relation, find the left and right pages. */
	Page		leftpg = NULL;
	BlockNumber	leftblk = InvalidBlockNumber;
	Page		rightpg = NULL;
	BlockNumber	rightblk = InvalidBlockNumber;
	BlockNumber nblocks = RelationGetNumberOfBlocks(indrel);
	BlockNumber blkno;
	BTPageOpaque leftopq;
	BTPageOpaque rightopq;
	OffsetNumber maxoff;
	OffsetNumber off;
	Page		pg;
	BTPageOpaque opq;
	double		split_ratio;
	StringInfoData sbuf;

	initStringInfo(&sbuf);

	/* scan all index pages to find the left and right page */
	for (blkno = 1; blkno < nblocks; blkno++)
	{
		Buffer buf = ReadBuffer(indrel, blkno);

		LockBuffer(buf, BUFFER_LOCK_EXCLUSIVE);

		pg = BufferGetPage(buf);
		opq = (BTPageOpaque) PageGetSpecialPointer(pg);

		if (opq->btpo.level == 0 && opq->btpo_prev == P_NONE)
		{
			leftblk = blkno;
			leftpg = PageGetTempPageCopy(pg);
		}
		if (opq->btpo.level == 0 && opq->btpo_next == P_NONE)
		{
			rightblk = blkno;
			rightpg = PageGetTempPageCopy(pg);
		}

		UnlockReleaseBuffer(buf);
	}

	if (!leftpg || !rightpg)
		elog(ERROR, "could not find page split halves");

	leftopq = (BTPageOpaque) PageGetSpecialPointer(leftpg);
	rightopq = (BTPageOpaque) PageGetSpecialPointer(rightpg);

	if (leftopq->btpo_next != rightblk ||
		rightopq->btpo_prev != leftblk)
		elog(ERROR, "unexpected next/prev pointers");

	/* Print all tuples */
	pg = leftpg;
	opq = leftopq;
	maxoff = PageGetMaxOffsetNumber(pg);
	for (off = P_FIRSTDATAKEY(opq); off <= maxoff; off++)
	{
		ItemId iid = PageGetItemId(pg, off);
		IndexTuple itup = (IndexTuple) PageGetItem(pg, iid);
 		bool	split_tuple;

		split_tuple = ItemPointerEquals(&itup->t_tid, &split_tid);
		
		if (off == P_FIRSTDATAKEY(opq) ||
			off == maxoff ||
			split_tuple)
		{
			appendStringInfo(&sbuf, "left  %3d/%3d: %s %s\n", off, maxoff,
							 print_index_tuple(indrel, itup),
							 split_tuple ? "SPLIT TUPLE" : "");
		}
	}

	pg = rightpg;
	opq = rightopq;
	maxoff = PageGetMaxOffsetNumber(pg);
	for (off = P_FIRSTDATAKEY(opq); off <= maxoff; off++)
	{
		ItemId iid = PageGetItemId(pg, off);
		IndexTuple itup = (IndexTuple) PageGetItem(pg, iid);
 		bool	split_tuple;

		split_tuple = ItemPointerEquals(&itup->t_tid, &split_tid);
		
		if (off == P_FIRSTDATAKEY(opq) ||
			off == maxoff ||
			split_tuple)
		{
			appendStringInfo(&sbuf, "right %3d/%3d: %s %s\n", off, maxoff,
							 print_index_tuple(indrel, itup),
							 split_tuple ? "SPLIT TUPLE" : "");
		}
	}

	split_ratio = (double) PageGetFreeSpace(rightpg) / (PageGetFreeSpace(leftpg) + PageGetFreeSpace(rightpg));
	appendStringInfo(&sbuf,"split ratio: %d/%d\n",
					 (int) round(split_ratio * 100),
					 (int) round(100 - split_ratio * 100));

	elog(NOTICE, "%s:\n%s", title, sbuf.data);
}

#define MAXROWS 1000

typedef char *(test_callback_fn) (Relation indrel, Relation heaprel, int tupno);

static void
run_split_test(char *title, test_callback_fn callback_fn)
{
	int			tupno;
	BlockNumber nblocks_before = 0;
	BlockNumber nblocks_after;
	ItemPointerData split_tid;
	char tblname[100];
	char idxname[100];
	StringInfoData sqlbuf;
	static int	counter = 0;
	Oid			indrelid;
	Oid			heapid;
	Relation	indrel;
	Relation	heaprel;

	counter++;
	snprintf(tblname, sizeof(tblname), "split_test_%d", counter);
	snprintf(idxname, sizeof(tblname), "split_test_%d_idx", counter);
	
	initStringInfo(&sqlbuf);
	appendStringInfo(&sqlbuf, "CREATE TEMPORARY TABLE %s (a int4, b int4) ON COMMIT DROP;\n", tblname);
	appendStringInfo(&sqlbuf, "CREATE INDEX %s ON pg_temp.%s(a, b);\n", idxname, tblname);
	SPI_execute(sqlbuf.data, false, 0);

	indrel = relation_openrv(makeRangeVar("pg_temp", idxname, -1), AccessExclusiveLock);
	indrelid = RelationGetRelid(indrel);

	heapid = IndexGetRelation(indrelid, false);
	heaprel = heap_open(heapid, AccessExclusiveLock);

	ItemPointerSetInvalid(&split_tid);
	
	for (tupno = 0; tupno < MAXROWS; tupno++)
	{
		ItemPointerData tid;

		char *tuple_str = callback_fn(indrel, heaprel, tupno);

		tid = insert_tuple(indrel, heaprel, tuple_str);

		if (tupno == 0)
		{
			nblocks_before = RelationGetNumberOfBlocks(indrel);
			if (nblocks_before != 2)
				elog(ERROR, "index has invalid size");
		}
		else
		{
			nblocks_after = RelationGetNumberOfBlocks(indrel);

			if (nblocks_after != nblocks_before)
			{
				split_tid = tid;
				break;
			}
		}
	}

	print_split(indrel, split_tid, title);
	
	index_close(indrel, AccessExclusiveLock);
	heap_close(heaprel, AccessExclusiveLock);
}

static char *
test_split_1(Relation indrel, Relation heaprel, int tupno)
{
	static char buf[100];

	snprintf(buf, sizeof(buf), "(1, %d)", tupno);

	return buf;
}


static char *
test_split_2(Relation indrel, Relation heaprel, int tupno)
{
	static char buf[100];

	snprintf(buf, sizeof(buf), "(%d, %d)", tupno, tupno);

	return buf;
}


static char *
test_split_3(Relation indrel, Relation heaprel, int tupno)
{
	static char buf[100];

	if (tupno < 100)
		snprintf(buf, sizeof(buf), "(%d, %d)", tupno, tupno);
	else
		snprintf(buf, sizeof(buf), "(10, 10)");

	return buf;
}


static char *
test_split_4(Relation indrel, Relation heaprel, int tupno)
{
	static char buf[100];

	if (tupno < 100)
		snprintf(buf, sizeof(buf), "(2, %d)", tupno);
	else
		snprintf(buf, sizeof(buf), "(1, %d)", tupno);

	return buf;
}

Datum
test_split(PG_FUNCTION_ARGS)
{
	int			ret;

	/* Connect to SPI manager */
	if ((ret = SPI_connect()) < 0)
		elog(ERROR, "SPI_connect returned %d", ret);

	run_split_test("test 1", test_split_1);
	run_split_test("test 2", test_split_2);
	run_split_test("test 3", test_split_3);
	run_split_test("test 4", test_split_4);

	SPI_finish();

	PG_RETURN_VOID();
}
